###########################################
## functions to compute FDR from empirically
###########################################

## generate negative bumps to calibrate FDR
negative.peaks <- function(data, distn, param, cfg) {
  data$signal <- -data$signal
  postprob <- gscan(data, distn, param$sita[2], param$q, cfg)
  ## peaks
  ncond <- dim(data$signal)[2]
  bumps.all <- bumps.single <- vector("list", ncond)
  for(i in 1:ncond) {
    bumps.single[[i]] <- peakFinder(postprob, data$allchr, data$allpos, cfg, cutoff=cfg$peak$Pvalue,
                                    cond=i, use.joint=FALSE, pi=param$pi1[i])
    bumps.all[[i]] <- peakFinder(postprob, data$allchr, data$allpos, cfg,
                                 cutoff=cfg$peak$Pvalue, cond=i, use.joint=TRUE)
  }
  if(!is.null(cfg$cel))
    names(bumps.single) <- names(bumps.all) <- names(cfg$cel)
  if(!is.null(cfg$group))
    names(bumps.single) <- names(bumps.all) <- names(cfg$cel)
  
  result <- list(single=bumps.single, pool=bumps.all, param=param)
  invisible(result)
}

## calculate FDR from negative peaks
calc.FDR <- function(peak1, peak0) {
  n0 <- nrow(peak1)
  fdr <- rep(0, n0)
  if(nrow(peak0)==0)
    return(fdr)
  
  counter <- 0
  for(i in 1:nrow(peak0)) {
    tmp <- sum(peak1$postprob[(counter+1):n0]>peak0[i, "postprob"])
    fdr[counter+(1:tmp)] <- (i-1)/(counter+tmp)
    counter <- counter+tmp
    if(counter == nrow(peak1))
      break
  }
  fdr
}
