###########################################################
## functions to filter the data to keep candidate regions
###########################################################
#################################################
## filter the data to get candidate region
## I'll do MA on t statistics and cut based FDR scores.
## Emission for two groups were estimated here.
#################################################
prefilter <- function(signal, vars, allchr, allpos, sep, peakLen, candidateLength,
                      nreps, fdr=0.3) {
##   if(!is.matrix(signal))
##     signal <- matrix(signal)
  ncond <- dim(signal)[2]
  nobs <- dim(signal)[1]
  
  ## init distn
  distn <- matrix(0, nrow=ncond, 4)
  colnames(distn) <- c("mu0", "mu1", "var0", "var1")
  
  if(length(vars) == 1) { ## no probe level variance, this is for single replicate.
    sds <- matrix(rep(1, ncond), nrow=1)
  } else {
    sds <- sqrt(vars)
  }
  flag=rep(FALSE, nobs)

  ##  smoothing window is from configuration
  sw <- peakLen
  ## compute number of probes in moving window
  nprobes <- .Call("nitem_bin", as.integer(allpos), as.integer(sw))

  ## compute t-statistics for each probe
  for(icond in 1:ncond) {
    outlier <- quantile(signal[,icond], 0.9999) ## cutoff for outlier
    ## deal with cases where some conditions have 1 reps
    single=FALSE
    if(all(sds[,icond]==0) | length(vars)==1)
      single=TRUE
    if(single)
      tstat <- signal[,icond]
    else
      tstat <- signal[,icond] / sds[,icond]
    ## moving average 
    res <- .Call("windowFilter_engine", as.double(tstat), nprobes)
    ## determine cutoff based on FDR
    cutoff <- findCutoff.fdr(res, fdr)
    idx.cond <- res>cutoff & signal[,icond]<outlier
    flag=flag | idx.cond
    ## mix in some backgrounds and do EM - this is tricky.
    ## EM mixture model is not stable!!!
    ## Maybe I should do a semi EM, with means estimated from data!!!
    idx.bg <- sample(which(!idx.cond), 2*sum(idx.cond, na.rm=TRUE), replace=TRUE)
    tmpidx <- idx.cond
    tmpidx[idx.bg] <- TRUE

    if(single) ## single replicate for this condition
       tmp <- mix.2norm(signal[tmpidx,icond], pi.grp2=0.2, var.eq=FALSE, verbose=FALSE)
    else ## multiple replicates
      tmp <- mix.2norm.var(signal[tmpidx,icond], vars[tmpidx,icond], pi.grp2=0.2, verbose=FALSE)
    distn[icond,] <- unlist(tmp[1:4])
  }
  ## indices for probes need to be kept
  allidx <- keepidx(flag, allchr, allpos, sep, candidateLength)
  ## Maybe I should do EM here on allidx???
  
  ## return
  list(idx=allidx, distn=distn)
}
  
  
###########################################################
## function to determine the cutoff values
###########################################################
findCutoff.fdr <- function(res, fdr) {
  foo <- function(p) {
    q <- quantile(res, p, na.rm=TRUE)
    n0 <- sum(res<q, na.rm=TRUE)
    n1 <- sum(res>-q, na.rm=TRUE)
    if(n1<1000) { ## too little items in H1
      return(100)
    } else {
      fdr <- n0/n1
      fdr
    }
  }
  fail <- FALSE
  ## divide and conquer??
  qq <- rep(0.001, 2)
  ## find correct upper and lower bounds
  fdr0 <- foo(qq[1])
  if(fdr0 > fdr) { ## initial guess is too big
    fdr.lo <- fdr0
    while(fdr.lo > fdr) { ## still bigger than desired FDR, need to go lower
      fdr.hi <- fdr.lo
      tmp <- qq[1]
      qq[2] <- qq[1]
      qq[1] <- tmp/2
      fdr.lo <- foo(qq[1])
      if(fdr.lo==100) {
        fail <- TRUE
        break
      }
    }
  } else { ## initial guess is too small
    fdr.hi <- fdr0
    while(fdr.hi < fdr) { ## still bigger than desired FDR, need to go lower
      fdr.lo <- fdr.hi
      tmp <- qq[2]
      qq[2] <- qq[2]+0.1
      qq[1] <- tmp
      fdr.hi <- foo(qq[2])
    }
  }

  ## couldn't find good cutoff, use arbitrary, keep top 0.5% or 2500 points (whichever is bigger)
  if(fail) {
    q <- 0.005
    if(q*length(res) < 2500) {
      q <- 2500 / length(res)
    }
    cutoff <- quantile(res, 1-q, na.rm=TRUE)
  } else {
    ## now boundries are set, do divide and conquer
    counter <- 0
    while((fdr.hi-fdr.lo)>0.01) {
      tmp <- mean(qq)
      fdr.tmp <- foo(tmp)
      if(fdr.tmp>fdr) {
        qq[2] <- tmp
        fdr.hi <- fdr.tmp
      } else {
        qq[1] <- tmp
        fdr.lo <- fdr.tmp
      }
      counter <- counter + 1
      if(counter > 50) {
        break
      }
    }
    cutoff <- abs(quantile(res, mean(qq), na.rm=TRUE))
  }
  
  cutoff
}


######################################################
## a new one, to find candidate peaks with fixed length
######################################################
keepidx <- function(flag, allchr, allpos, sep, candidateLength) {
  allidx <- rep(FALSE, length(flag))
  ## by chr
  idx <- split(1:length(flag), allchr)
  for(ichr in seq(along=idx)) {
    idx.chr <- idx[[ichr]]
    tmp <- .Call("findCandidate", as.integer(flag[idx.chr]), as.integer(allpos[idx.chr]),
                 as.integer(sep), as.integer(candidateLength))
    allidx[idx.chr] <- tmp
  }
  which(allidx==1)
  ## I will skip merging, 'cause this is only for estimating parameters so it's not important
}


