## scan the genome to find peaks with given parameters.
## output is a matrix of posterior probabilities

gscan <- function(data, distn, pi, q, cfg) {
  nprobe <- dim(data$signal)[1]
  ncond <- dim(data$signal)[2]
  ## compute densities and lr
  dens <- compute.density(data$signal, data$vars, distn)
  ylr <- dens$lden1 - dens$lden0
  ## do it by chr
  idx <- split(1:nprobe, data$allchr)
  nchr <- length(idx)
  peaklr <- nprobes <- res <- vector("list", nchr)

  for(ichr in 1:nchr) {
    idx.chr <- idx[[ichr]]
    tmp <- .Call("scan_engine", ylr[idx.chr,,drop=FALSE], as.integer(data$allpos[idx.chr]),
                 as.double(pi), as.double(q), as.integer(cfg$peak$bumpLength),
                 as.integer(cfg$peak$candidateLength),as.integer(cfg$peak$maxGap))
    res[[ichr]] <- tmp[[1]]
    nprobes[[ichr]] <- tmp[[2]]
    peaklr[[ichr]] <- tmp[[3]]
    ## finding peaks for this chr
  }
  names(nprobes)=names(res)=names(peaklr)=names(idx)
  invisible(list(postprob=res, ylr=ylr, nprobes=nprobes, lr=peaklr))
}

######################################################################
## find peaks based on results from gscan
## the input pi is the probability of starting a peak.
## It'll be used in making single condition peaks.
######################################################################
peakFinder <- function(postprob, allchr, allpos, cfg, cutoff=0.1, cond=1, use.joint=TRUE, pi) {
  nprobe <- length(allchr)
  ncond <- dim(postprob[[1]][[1]])[2] - 1
  idx <- split(1:nprobe, allchr)
  nchr <- length(idx)
  ## init results
  n0 <- 100000
  peaks <- data.frame(chr=rep("",n0), start=rep(0,n0), end=rep(0,n0), peak=rep(0, n0), length=rep(0,n0),
                      nprobes=rep(0,n0), lr=rep(0,n0), postprob=rep(0,n0))
  levels(peaks$chr) <- names(idx)
  counter <- 0
  for(ichr in 1:nchr) {
    thislr <- postprob$lr[[ichr]][,cond] ## this is the lr of starting a peak at this location
    pp <- postprob$postprob[[ichr]][,cond]
    pp[length(pp)] <- 0
    if(use.joint)  ## use all conditions, multiply positional priors
      pp <- pp * postprob$postprob[[ichr]][,ncond+1]
    ## find candidate regions
    thispos <- allpos[idx[[ichr]]]
    this.ylr <- postprob$ylr[idx[[ichr]],cond] ## this is the lr in probe level
    ## scan the posterior probabilities and find locations for peaks
    tmp <- .Call("findBumps", as.integer(thispos), as.double(pp),as.double(thislr),
                 as.double(cutoff), as.integer(postprob$nprobes[[ichr]]),
                 as.integer(cfg$peak$bumpLength), as.double(this.ylr))
    if(length(tmp)==0)
      next
    tmp <- matrix(tmp, ncol=4, byrow=TRUE) + 1
    peakloc <- tmp[,1]
    ## probabilities for peak
    if(use.joint) 
      this.postprob <- pp[peakloc] 
    else { ## recompute postprob for single condition because of different priors
      dtmp <- sapply(thislr[peakloc], function(x) Raddlog(x+log(pi), log(1-pi)))
      this.postprob <- exp(thislr[peakloc]+log(pi)-dtmp)
    }
    tmppeak <- data.frame(chr=names(idx)[ichr], start=thispos[tmp[,2]], end=thispos[tmp[,3]],
                          peak=thispos[tmp[,4]], length=thispos[tmp[,3]]-thispos[tmp[,2]],
                          nprobes=tmp[,3]-tmp[,2],lr=thislr[peakloc], postprob=this.postprob)
    peaks[counter+(1:nrow(tmppeak)),] <- tmppeak
    counter <- counter + nrow(tmppeak)
  }
  ## remove the really short ones
  peaks <- peaks[peaks$length>(cfg$peak$bumpLength[1]/2),]
  peaks <- peaks[peaks$nprobes>=cfg$peak$MinProbe,]
  ## sort by posterior probabilities and break ties by lr
  ix <- rev(do.call("order", peaks[,c("postprob","lr")])) # how to use decreasing=TRUE in do.call??
  peaks <- peaks[ix,]
  ## extend both ends by 50bps
  peaks$start <- peaks$start-50
  peaks$end <- peaks$end+50
  
  peaks
  
}

