## Functions to read in configuration file and data
## To do:
## read.CEL some have single replicates and others have multiple replicates.
## test

#######################################################
## This function will read in the configuration file
## and return a list.
#######################################################
read.config <- function(cfgfile) {
  a=readLines(cfgfile)
  for(j in 1:length(a)) {
    ## if user input tab/comma/, convert to space
    a[j]=gsub("\t", " ", a[j])
    a[j]=gsub(",", " ", a[j])
  }
  
  ## remove empty lines and lines start with #
  idx.empty=nchar(a)==0
  idx.comment=substr(a,1,1)=="#"
  idx.rm=idx.empty | idx.comment
  a=a[!idx.rm]

  config=NULL
  Group=NULL
  ## parse a
  isec=which(substr(a,1,1)=="[")
  lines.sec=cbind(isec, c(isec[-1]-1,length(a)))
  for(i in 1:length(isec)) { ## parse each section
    n=lines.sec[i,2] - lines.sec[i,1] 
    title=a[lines.sec[i,1]]
    title=substr(title, 2, nchar(title)-1)
    if(title=="data") {
      data=vector("list", n)
      counter=1
      for(j in (lines.sec[i,1]+1):lines.sec[i,2]) {
        tmp=strsplit(a[j], "=")[[1]]
        data[[counter]]=sub(" ", "", tmp[2])
        names(data)[counter]=sub(" ", "", tmp[1])
        counter=counter+1
      }
      if(is.null(data$Format)) ## missing format
        data$Format <- "cel" ## inputs are CEL files by default
      if(data$Format=="cel")
        data$Pair=as.integer(data$Pair) ## whether cel files are paired
    }
      
    else if(title=="cel") {
      cel=vector("list", n)
      counter=1
      for(j in (lines.sec[i,1]+1):lines.sec[i,2]) {
        tmp=strsplit(a[j], "=")[[1]]
        tmp2=strsplit(tmp[2], " ")[[1]]
        ## remove empty entries (caused by extra space ,etc.)
        tmp2=tmp2[nchar(tmp2)>0]
        cel[[counter]]=tmp2
        names(cel)[counter]=sub(" ", "", tmp[1])
        counter=counter+1
      }
    }
    else if(title=="Group") { ## read group labels
      Group=vector("list", n)
      counter=1
      for(j in (lines.sec[i,1]+1):lines.sec[i,2]) {
        tmp=strsplit(a[j], "=")[[1]]
        tmp2=tmp[2]
        Group[[counter]]=as.integer(strsplit(tmp2,"")[[1]])
        names(Group)[counter]=sub(" ", "", tmp[1])
        counter=counter+1
      }
    }
    else if(title=="Condition") { ## read column numbers for conditions, this is when input is text
      Group=vector("list", n)
      counter=1
      for(j in (lines.sec[i,1]+1):lines.sec[i,2]) {
        tmp=strsplit(a[j], "=")[[1]]
        tmp2=tmp[2]
        Group[[counter]]=as.integer(strsplit(tmp2," ")[[1]])
        names(Group)[counter]=sub(" ", "", tmp[1])
        counter=counter+1
      }
    }
    
    else if(title=="peak finding") { ## for peaking finding
      peak=vector("list", n)
      counter=1
      for(j in (lines.sec[i,1]+1):lines.sec[i,2]) {
        tmp=strsplit(a[j], "=")[[1]]
        peak[[counter]]=tmp[2]
        names(peak)[counter]=sub(" ", "", tmp[1])
        counter=counter+1
      }
      peak$MinProbe=as.integer(peak$MinProbe)
      peak$bumpLength=as.integer(strsplit(peak$bumpLength, " ")[[1]])
      peak$candidateLength=as.integer(peak$candidateLength)
      peak$maxGap=as.integer(peak$maxGap)

      if(!is.null(peak$Pvalue))
        peak$Pvalue=as.double(peak$Pvalue)

      if(!is.null(peak$FDRcutoff))
        peak$FDRcutoff=as.double(peak$FDRcutoff)
      
      if(is.null(peak$computeFDR))
        peak$computeFDR=0
      else
        peak$computeFDR=as.integer(peak$computeFDR)
    }
    
  }

  if(data$Format=="cel") {
    for(i in 1:length(cel)) {
      for(j in 1:length(cel[[i]])) {
        cel[[i]][[j]] <- paste(data$CelFolder, cel[[i]][[j]], sep="/")
      }
    }
  }
  
  config$data=data
  config$peak=peak
  if(data$Format=="cel") {
    config$cel=cel
    if(data$Pair==0) { ## not paired, read in group label
      if(is.null(Group))
        stop("Must provide group labels in configuration file for non-paired experiment.\n")
      config$group=Group
    }
  }

  if(data$Format=="text") { ## when input is text, take group column numbers
    config$group=Group
  }

  config
}


#######################################################
## Given configuration do
## 1. Read in CEL files
## 2. Preprocess (quantile normalization).
## 3. Compute mean and variances.
## 4. save data into rda file
## 5. Return signal and variances
#######################################################
read.data.CEL <- function(cfg) {
  ## read in bpmap
  cat("reading bpmap file ...\n")
  bpmap <- readBpmap(cfg$data$Bpmap, readProbeSeq=FALSE, readSeqInfo=FALSE,
                     readMMXY=FALSE, readStrand=FALSE)
  ## remove duplicated probes
  for(ichr in 1:length(bpmap)) {
    flag <- !duplicated(bpmap[[ichr]]$startpos)
    bpmap[[ichr]]$startpos <- bpmap[[ichr]]$startpos[flag]
    bpmap[[ichr]]$pmx <- bpmap[[ichr]]$pmx[flag]
    bpmap[[ichr]]$pmy <- bpmap[[ichr]]$pmy[flag]
  }

  ## read in a CEL header
  cel.header <- readCelHeader(cfg$cel[[1]][1])
  cel.indices <- lapply(bpmap, function(x) xy2indices(x[["pmx"]],x[["pmy"]], nr = cel.header[["cols"]]))
  allidx <- unlist(cel.indices)
  allpos <- as.integer(unlist(lapply(bpmap, function(x) x$startpos))) ## positions are integers
  allchr <- rep(names(cel.indices), unlist(lapply(cel.indices, length)))
  
  ## 1. read in all CEL files
  cat("reading CEL files ...\n")
  ncond <- length(cfg$cel)
  y <- log2(readCelIntensities(unlist(cfg$cel), allidx))
  
  ## 2. quantile normalization
  cat("Quantile normalization ... \n")
  if(ncond>1) {
    if(cfg$data$Pair == 1) { ## paired data,
      ## use average of control channels to be target
      target <- sort(rowMeans(y[,seq(2,dim(y)[2],2)]))
    }
    else { ## not paired, use all control arrays
      ## make index for ip/control columns in y
      idx.ip <- idx.ct <- NULL
      counter <- 0
      for(icond in 1:length(cfg$cel)) {
        group.cond <- cfg$group[[names(cfg$cel[icond])]]
        idx.ip <- c(idx.ip, counter+which(group.cond==1))
        idx.ct <-c(idx.ct, counter+which(group.cond==0))
        counter <- counter + length(group.cond)
      }
      target <- sort(rowMeans(y[,idx.ip]))
    }
  } else { ## use saved target
    data(target)
    target <- sort(sample(log2(target), length(allchr), replace=TRUE))
  }
  for(i in 1:ncol(y)) {
    tmpsort <- sort(y[,i], index.return=TRUE)
    y[tmpsort$ix,i] <- target
  }

  ## 3. compute probe level signals and variances in each condition
  if(cfg$data$Pair == 1) { ## paired experiment
    ## first compute all log ratios - note for paired data, y should be in IP1, CT1, IP2, CT2, ...
    n <- ncol(y)
    logratio <- y[,seq(1,n,by=2)] - y[,seq(2,n,by=2)]
    ## number of replicates for each condition
    nreps <- unlist(lapply(cfg$cel, length))/2
    ## compute signal and probe level variances
    tmp <- calc.MeanVar.paired(logratio, nreps)
    signal <- tmp$signal
    vars <- tmp$vars
  } else { ## unparied experiment
    tmp <- calc.MeanVar.unpaired(y, cfg$group)
    signal <- tmp$signal
    vars <- tmp$vars
  }

  ## there are more cases I need to take care of:
  ## 1. Some conditions are paired and some are not
  
  ## sort data according to chr and pos
  tmp <- (max(allpos)+1)*as.integer(as.factor(allchr)) + allpos
  ix <- sort(tmp, index.return=TRUE)$ix
  signal=signal[ix,,drop=FALSE]
  if(length(vars)>1)
    vars=vars[ix,,drop=FALSE]
  allchr=allchr[ix]
  allpos=allpos[ix]

  ## Save data
  data <- list(signal=signal, vars=vars, allchr=allchr, allpos=allpos)
  fn <- paste(cfg$data$WorkFolder, "/", cfg$data$Title, "-data.rda", sep="")
  save(data, file=fn)
  data
}


##################################################
## read in file when it's in text
## The entries in text should be preproessed log ratios
## so I'll skip normalization
## I'll pick regions and run EM
## Need to modify this for unequal number of replicates !!!
##################################################
read.data.text <- function(cfg) {
  tmp <- read.table(cfg$data$file, sep="\t")
  allchr <- tmp[,1]
  ## modify if chr are integers
  if(is.integer(allchr))
    allchr <- paste("chr", allchr, sep="")
  allpos <- as.integer(tmp[,2]) ## positions are integers
  truncate.flag <- FALSE ## apply truncation on variances

  ## compute signal and variances
  yy <- as.matrix(tmp[,unlist(cfg$group)])
  nreps <- unlist(lapply(cfg$group, length))
  tmp <- calc.MeanVar.paired(yy, nreps)
  signal <- tmp$signal
  vars <- tmp$vars
  
  ## sort data according to chr and pos
  tmp <- (max(allpos)+1)*as.integer(as.factor(allchr)) + allpos
    ix <- sort(tmp, index.return=TRUE)$ix
  signal=signal[ix,,drop=FALSE]
  if(length(vars)>1)
    vars=vars[ix,,drop=FALSE]
  allchr=allchr[ix]
  allpos=allpos[ix]
  ## Save all the data
  data <- list(signal=signal, vars=vars, allchr=allchr, allpos=allpos,nreps=nreps)
  fn <- paste(cfg$data$WorkFolder, "/", cfg$data$Title, "-data.rda", sep="")
  save(data, file=fn)
  data
}

###########################################################
## function to compute probe level signals and variances.
## yy: log ratio
## nreps: Number of replicates for each condition
###########################################################
calc.MeanVar.paired <- function(yy, nreps) {
  ncond <- length(nreps)
  nprobe <- nrow(yy)
  signal <- vars <- matrix(0, nrow=nprobe, ncol=ncond)
  
  if(all(nreps>1)) { ## all with multiple replicates
    counter <- 0
    for(icond in 1:ncond) { ## loop for condition
      tmp <- yy[,counter+(1:nreps[icond])]
      signal[,icond] <- rowMeans(tmp)
      vars[,icond] <- rowSums((tmp-signal[,icond])^2)/(nreps[icond]-1)/nreps[icond]
      vars[,icond] <- ji.shrinker(vars[,icond], nreps[icond]-1)
      counter <- counter + nreps[icond]
    }
  } else if(all(nreps==1)) { ## all with single replicates
    signal <- yy
    vv <- rowVars(signal)
    tmp <- quantile(vv, 0.9)
    truncate.flag <- TRUE
    vv[vv>tmp] <- tmp
    tmp <- ji.shrinker(vv, (ncond-1), 0.9)
    for(i in 1:ncol(vars))
      vars[,i] <- tmp
  } else { ## some have single replicate, some have more.
    counter <- 0
    SSE <- 0
    for(icond in 1:ncond) { ## loop for condition
      tmp <- yy[,counter+(1:nreps[icond])]
      if(nreps[icond]>1) {
        signal[,icond] <- rowMeans(tmp)
        tmp2 <- rowSums((tmp-signal[,icond])^2)
        SSE <- SSE + tmp2
        vars[,icond] <- tmp2/(nreps[icond]-1)/nreps[icond]
        vars[,icond] <- ji.shrinker(vars[,icond], nreps[icond]-1)
      } else
      signal[,icond] <- tmp
      counter <- counter + nreps[icond]
    }
    ## for conditions with single replicate, use pooled variances from conditions with more replicate
    v <- sum(nreps-1)
    var.pooled <- SSE/v
    var.pooled <- ji.shrinker(var.pooled, v)
    vars[,nreps==1] <- var.pooled
  }

  colnames(signal) <- colnames(vars) <- names(nreps)
  list(signal=signal, vars=vars)
}

    
###########################################################
## function to compute probe level signals and variances.
## yy: log intensity
## group: labels for IP and control for each condition
###########################################################
calc.MeanVar.unpaired <- function(yy, group) {
  nprobe <- nrow(yy)
  ncond <- length(group)
  ## see whether there are replicates in each condition
  reps.flag <- unlist(lapply(group, function(x) length(x)>3))
  signal <- vars <- matrix(0, nrow=nprobe, ncol=ncond)
  
  if(all(reps.flag)) { ## all with multiple replicates
    counter <- 0
    for(icond in 1:ncond) {
      group.cond <- group[[icond]]
      idx.ip <- counter+which(group.cond==1); nip <- length(idx.ip)
      idx.ct <- counter+which(group.cond==0); nct <- length(idx.ct)
      if(nip==0)
        stop("No IP array in unpaired experiment.\n")
      if(nct==0)
        stop("No control array in unpaired experiment.\n")

      ip <- yy[,idx.ip]
      if(nip > 1) {
        ip.mean <- rowMeans(ip)
        SSW.ip <- rowSums((ip-ip.mean)^2)
      } else {
        ip.mean <- ip
        SSW.ip <- 0
      }

      ct <- yy[,idx.ct]
      if(nct>1) { ## control channel
        ct.mean <- rowMeans(ct)
        SSW.ct <- rowSums((ct-ct.mean)^2)
      } else {
        ct.mean <- ct
        SSW.ct <- 0
      }

      ## compute signals and variances
      signal[,icond] <- ip.mean - ct.mean
      v <- nip + nct - 2
      vars[,icond] <- (SSW.ip+SSW.ct)/v/(1/nip+1/nct)
      vars[,icond] <- ji.shrinker(vars[,icond], v)
      ## increment counter and go to next condition
      counter <- counter + length(group.cond)
    }
  } else if(all(!reps.flag)) { ## all with single replicate (one IP, one control)
    ## first compute all signals
    counter <- 0
    for(icond in 1:ncond) {
      group.cond <- group[[icond]]
      idx.ip <- counter+which(group.cond==1)
      idx.ct <- counter+which(group.cond==0)
      signal[,icond] <- yy[,idx.ip] - yy[,idx.ct]
      counter <- counter + 2
    }
    ## use pooled variance - maybe should just use control channel
    ## but in many cases they use a common control
    ## I'm using all intensity with robust procedure now. 
    Si <- 2*rowVars(yy)
    tmp <- quantile(Si, 0.9)
    Si[Si>tmp] <- tmp
    Si <- ji.shrinker(Si, ncond-1, 0.9)
    for(icond in 1:ncond) 
      vars[,icond] <- Si
  }    else { # some have single replicate, some have more.
    SSE <- totalV <- counter <- 0
    for(icond in 1:ncond) {
      group.cond <- group[[icond]]
      idx.ip <- counter+which(group.cond==1); nip <- length(idx.ip)
      idx.ct <- counter+which(group.cond==0); nct <- length(idx.ct)
      if(nip==0)
        stop("No IP array in unpaired experiment.\n")
      if(nct==0)
        stop("No control array in unpaired experiment.\n")
      
      ip <- yy[,idx.ip]
      if(nip > 1) {
        ip.mean <- rowMeans(ip)
        SSW.ip <- rowSums((ip-ip.mean)^2)
      } else {
        ip.mean <- ip
        SSW.ip <- 0
      }
      
      ct <- yy[,idx.ct]
      if(nct>1) { ## control channel
        ct.mean <- rowMeans(ct)
        SSW.ct <- rowSums((ct-ct.mean)^2)
      } else {
        ct.mean <- ct
        SSW.ct <- 0
      }
      
      ## compute signals and variances
      signal[,icond] <- ip.mean - ct.mean
      v <- nip + nct - 2
      if(v>0) {
        vars[,icond] <- (SSW.ip+SSW.ct)/v/(1/nip+1/nct)
        vars[,icond] <- ji.shrinker(vars[,icond], v)
        SSE <- SSE + SSW.ip + SSW.ct
        totalV <- totalV + v
      }
      ## increment counter and go to next condition
      counter <- counter + length(group.cond)
    }
    ## compute variances for those with single replicate
    Si <- 2 * SSE / totalV
    Si <- ji.shrinker(Si, totalV)
    vars[,which(!reps.flag)] <- Si
  }

  ## return
  colnames(signal) <- colnames(vars) <- names(group)
  list(signal=signal, vars=vars)
}
