/* function to find bumps, R caller is:
   ii <- .Call("findBumps", pos[idx[[ichr]]], x[idx[[ichr]]], cutoff, as.integer(minlen),
               as.integer(mincount), dis.merge, as.integer(trim))
*/

#include "jamie.h"
#include "util.h"

SEXP findBumps(SEXP R_pos, SEXP R_x, SEXP R_lr, SEXP R_cutoff,SEXP R_nprobes, SEXP R_bumpLen, SEXP R_ylr);
void bestPeak(double *y, int *pos, int nprobes, int *bumplen, int ntype,
              double baseline, int startidx, int *result);

SEXP findBumps(SEXP R_pos, SEXP R_x, SEXP R_lr, SEXP R_cutoff,SEXP R_nprobes, SEXP R_bumpLen, SEXP R_ylr) {
  int i, ipos, nobs, nbumptype, n0=100000, bumpflag, inbump, idxMax, tmpstart, tmpend;
  int *tmpidx, *pos, *nprobes, *bumpLen, counter=0;
  double tmplen, dtmp1, dtmp2, *x, *lr, *ylr, cutoff, *ylr_cumsums, baseline;

  nobs=length(R_x);
  nbumptype=length(R_bumpLen);
  pos=INTEGER(R_pos);
  nprobes=INTEGER(R_nprobes);
  bumpLen=INTEGER(R_bumpLen);
  x=REAL(R_x);
  cutoff=*REAL(R_cutoff);
  lr=REAL(R_lr);
  ylr=REAL(R_ylr);

  tmpidx=(int *) R_alloc(n0*4, sizeof(int));

  /* compute cumsum - this will save some computation */
  ylr_cumsums=(double *) R_alloc(nobs, sizeof(double));
  cumsum1(ylr, nobs, ylr_cumsums);

  /* scan the vector to get start/end positions */
  bumpflag = x[0]>cutoff;
  if(bumpflag) {
    inbump = 1;
    tmpidx[0]=0;
  }

  ipos=0;
  while(ipos<nobs) {
    bumpflag=x[ipos]>cutoff;
    if(bumpflag) {     /* in a bump, find local maxima within candLen bp, use lr to break ties (postprob=1) */
      idxMax=0;
      dtmp1=x[ipos];
      dtmp2=lr[ipos];
      for(i=1; i<nprobes[ipos]; i++) {
	if(x[ipos+i]>=dtmp1 & lr[ipos+i]>dtmp2 ) {
	  idxMax = i;
	  dtmp1 = x[ipos+i];
	  dtmp2=lr[ipos+i];
	}
      }
      tmpstart=ipos+idxMax;
      tmpend=tmpstart+nprobes[ipos+idxMax]-1;
      tmpidx[counter*4]=tmpstart;
      /* locate the best regions for peak, write results to tmpidx */
      if(tmpstart==0)
	baseline=0;
      else
	baseline=ylr_cumsums[tmpstart-1];
      bestPeak(ylr_cumsums+tmpstart, pos+tmpstart, nprobes[ipos+idxMax], bumpLen, 
	       nbumptype, baseline, tmpstart, tmpidx+counter*4+1);
      ipos = tmpend + 1;
      counter ++;
    } else
      ipos ++;
  }
  /* return */
  SEXP res;
  PROTECT(res = allocVector(INTSXP, counter*4));
  memcpy(INTEGER(res), tmpidx, counter*4*sizeof(int));
  UNPROTECT(1);
  return(res);
}

/* search for boundary for best peak */
void bestPeak(double *y, int *pos, int nprobes, int *bumplen, int ntype, 
	      double baseline, int startidx, int *result) {
  double dtmp=0, dbest=-10000.0;
  int i, j, k, tmpstart, tmpend, ipeak;

  for(i=0; i<ntype; i++) { /* loop on different bump lengths */
    for(j=0; j<nprobes-3; j++) { /* different starting probe */
      for(k=j+3; k<nprobes; k++) { /* find ending probe */
        if(pos[k]-pos[j]>bumplen[i]) { /* peak is from j to k */
          /* this is a peak of given length, compute the lr */
          if(j==0)
            dtmp = y[k-1]-baseline;
          else
            dtmp = y[k-1] - y[j-1];
	  if(dtmp>dbest) { /* best boundary */
	    tmpstart=j;
	    tmpend=k;
	    dbest=dtmp;
	  }
	  break;
        }
      }
    }
  }
  /* now optimal boundary is located, find peak location by doing moving average */
  dbest = -10000.0;
  ipeak=0;
  for(i=tmpstart; i<=tmpend-3; i++) {
    if(i==0)
      dtmp = y[i+2]-baseline;
    else
      dtmp = y[i+2]-y[i-1];
    if(dtmp>dbest) {
      dbest=dtmp;
      ipeak=i+1;
    }
  }
  /* assign results, extend both ends by two probes */
  result[0]=startidx+tmpstart;
  result[1]=startidx+tmpend;
  result[2]=startidx+ipeak;
}
