/* caller is: tmp <- .Call("findCandidate", flag[idx.chr], allpos[idx.chr], sep, candBumpLen)*/
#include "jamie.h"
#include "util.h"
#include "gscan.h"
#define THRESH 200.0


SEXP findCandidate(SEXP R_flag, SEXP R_allpos, SEXP R_sep, SEXP R_candBumpLen) {
  int *flag, *allpos, sep, candBumpLen, ipos, nobs, inbump, tmpstart=0, tmpend=0, i, 
    bumpCenter, centerPos, tmpdis, tmplen, *result;
  SEXP R_result;

  flag=INTEGER(R_flag);
  allpos=INTEGER(R_allpos);
  sep=*INTEGER(R_sep);
  candBumpLen=*INTEGER(R_candBumpLen)/2;
  nobs=length(R_flag);

  PROTECT(R_result = allocVector(INTSXP, nobs));
  result = INTEGER(R_result);
  for(i=0; i<nobs; i++)
    result[i] = 0;

  inbump=flag[0];
  for(ipos=0; ipos<nobs; ipos++) {
    if(inbump) {     /* if there's open bump */
      if( (allpos[ipos]-allpos[ipos-1])>sep | flag[ipos]==0 ) {
        /* either hit the end of region or down edge, need to close this bump */
        inbump = 0;
        tmpend=ipos-1;
	makeFixedLen(allpos, tmpstart, tmpend, candBumpLen, nobs, result);
      }
    }
    else { /* not in bump region */
      if(flag[ipos]) {
	inbump = 1;
	tmpstart = ipos;
      }
    }
  }

  /* careful at the end of chain */
  if(inbump) { /* need to close the bump */
    tmpend = nobs-1;
    makeFixedLen(allpos, tmpstart, tmpend, candBumpLen, nobs, result);
  }

  UNPROTECT(1);
  return(R_result);
}


void makeFixedLen(int *allpos, int tmpstart, int tmpend, int candBumpLen, int nobs, int *result) {
  int bumpCenter, centerPos, i, tmpdis, tmplen;
  
  /* need to extend/shrink boths to make it fixed length */
  bumpCenter=(tmpstart+tmpend)/2;
  centerPos=allpos[bumpCenter];
  for(i=bumpCenter-1; i>=0; i--) { /* left side */
    tmpdis = centerPos - allpos[i];
    if(tmpdis>candBumpLen) {
      tmpstart = i+1;
      break;
    }
  }
  tmplen = 2*candBumpLen-(centerPos-allpos[tmpstart]);
  for(i=bumpCenter+1; i<nobs; i++) { /* rigth side */
    tmpdis = allpos[i] - centerPos;
    if(tmpdis>tmplen) {
      tmpend = i-1;
      break;
    }
  }
  /* assign if long enough */
  if(tmpend-tmpstart>=3)
    for(i=tmpstart; i<=tmpend; i++) 
      result[i] = 1;
}

/* compute the likelihood ratio for being a peak for a candidate region */
SEXP calc_Candidate_Lr(SEXP R_ylr, SEXP R_allpos, SEXP R_sep, SEXP R_bumplen) {
  double *ylr, *result, *ylr_cumsums, baseline;
  int *bumplen, nbumptype, sep, *allpos, i, j, counter=0, n0=10000, nobs, ncond, *itmp,
    tmpstart=0, tmpend, tmpn, len;
  SEXP R_result;

  ylr=REAL(R_ylr);
  allpos=INTEGER(R_allpos);
  sep=*INTEGER(R_sep);
  bumplen=INTEGER(R_bumplen);
  nbumptype=length(R_bumplen);
  itmp=INTEGER(getAttrib(R_ylr, R_DimSymbol));
  nobs=itmp[0];
  ncond=itmp[1];
  n0=n0*ncond;

  /* compute cumsum - this will save some computation */
  ylr_cumsums=(double *) R_alloc(nobs*ncond, sizeof(double));
  cumsum2(ylr, nobs,  ncond, ylr_cumsums);
  result = (double *)R_alloc(n0*ncond, sizeof(double));

  counter = 0;
  for(i=1; i<nobs-1; i++) { 
    /* compute for each candidate region */
    if(allpos[i]-allpos[i-1]>sep | i==nobs-1) { 
      if(i==nobs-1)
	tmpend=i;
      else 
	tmpend = i-1;
      for(j=0; j<ncond;j++) {
	if(tmpstart==0)
	  baseline=0;
	else 
	  baseline=ylr_cumsums[j*nobs+tmpstart-1];
	len = tmpend-tmpstart+1;
	if(len<3)
	  result[counter*ncond+j] = -1000;
	else if (len==3) 
	  result[counter*ncond+j] = dsum(ylr_cumsums+j*nobs+tmpstart, 3);
	else 
	  result[counter*ncond+j]=calcOnelr(ylr_cumsums+j*nobs+tmpstart, allpos+tmpstart, 
					    tmpend-tmpstart+1, bumplen, nbumptype, baseline);
      }
      counter++;
      tmpstart = i;
    }
  }

  PROTECT(R_result = allocVector(REALSXP, counter*ncond));
  memcpy(REAL(R_result), result, counter*ncond*sizeof(double));
  UNPROTECT(1);
  return(R_result);
}

/* compute lr for one candidate peak. Note input y is cumsum */
double calcOnelr(double *y, int *pos, int len, int *bumplen, int ntype, double baseline) {
  double res=-300, dtmp=0;
  int i, j, k, counter=0, minlen, flag;

  for(i=0; i<ntype; i++) { /* loop on different bump lengths */
    /* minimum length for a peak */
    if(i==0) 
      minlen = imax2(bumplen[0]-100, bumplen[0]/2);
    else 
      minlen = bumplen[i-1];
    
    for(j=0; j<len-2; j++) {
      flag = 0;
      for(k=j+3; k<len; k++) {
	if(pos[k]-pos[j]>bumplen[i]) { /* peak is from j to k-1 */
	  if(pos[k-1]-pos[j] < minlen) 
	    /* too short, can't find peak starting from j and satisfying the length */
	    break;
	  /* this is a peak of given length, compute the lr */
	  if(j==0)
	    dtmp = y[k-1]-baseline;
	  else
	    dtmp = y[k-1] - y[j-1];
	  res = logspace_add(res, dtmp);
	  counter ++;
	  flag = 1;
	  break;
	}
      }
      /* above didn't consider the last probe in the region. If we failed to
	 find any peak, deal with the last probe. */
      if( (len-j>=3) & (pos[len-1]-pos[j]>minlen) ) {
	/* this is a peak of given length, compute the lr */
	if(j==0)
	  dtmp = y[len-1]-baseline;
	else
	  dtmp = y[len-1] - y[j-1];
	res = logspace_add(res, dtmp);
	counter ++;
      }
    }
  }
  if(counter>0)
    return(res-log(counter));
  else 
    return(-1000); 
}


