/* utility functions for filtering */
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <R.h>
#include <Rmath.h>
#include <R_ext/PrtUtil.h>
#include <Rinternals.h>
#include <Rdefines.h>
#include <R_ext/RConverters.h>
#include <R_ext/Rdynload.h>

SEXP nitem_bin(SEXP pos_sexp, SEXP ws_sexp);
SEXP windowFilter_engine(SEXP x, SEXP nitem);
SEXP windowFilter(SEXP x, SEXP pos, SEXP ws);


/* find number of items in each bin.
   result is a two column matrix for number of items
   before and after the current position.
   I'll take a maximum of 11 probes on each side.  */
SEXP nitem_bin(SEXP pos_sexp, SEXP ws_sexp) {
  int ipos, j, *result_ptr, n;
  SEXP result;
  int npos = length(pos_sexp);
  int *pos = INTEGER(pos_sexp);
  int ws = *INTEGER(ws_sexp);
  PROTECT(result = allocMatrix(INTSXP, npos, 2));
  result_ptr = INTEGER(result);

  for(ipos=0; ipos<npos; ipos++) {
    /* left side */
    n = 0;
    for(j=ipos-1; j>=fmax(ipos-11,0); j--) {
      if(pos[ipos]-pos[j]>ws) {
	n = ipos - j - 1; 
	break;
      }
      n = 11;
    }
    if(j == -1) /* begining */
      n = ipos; 
    result_ptr[ipos] = n;

    /* right side */
    n = 0;
    for(j=ipos+1; j<fmin(ipos+11,npos); j++) {
      if(pos[j]-pos[ipos]>ws) {
        n = j-ipos-1;
	break;
      } 
      n = 11;
    }
    if(j == npos) /* end */
      n = npos - ipos - 1;
    result_ptr[ipos+npos] = n;
  }
  UNPROTECT(1);
  return(result);
}


/* engine function for moving avearge */
SEXP windowFilter_engine(SEXP x_sexp, SEXP nitem_sexp) {
  int ipos, j;
  double s;
  SEXP result;
  int npos = length(x_sexp);
  double *x = REAL(x_sexp);
  int *nitem = INTEGER(nitem_sexp);
  PROTECT(result = allocVector(REALSXP, npos));
  double *result_ptr = REAL(result);

  for(ipos=0; ipos<npos; ipos++) {
    s = 0;
    for(j=ipos-nitem[ipos]; j<=ipos+nitem[ipos+npos]; j++) 
      s = s + x[j];
    result_ptr[ipos] = s/(nitem[ipos]+nitem[ipos+npos]+1);
  }
  UNPROTECT(1);
  return(result);
}

/* wrapper function to do moving average within a fixed window size */
SEXP windowFilter(SEXP x, SEXP pos, SEXP ws) {
  SEXP nitem = nitem_bin(pos, ws);
  SEXP result = windowFilter_engine(x, nitem);
  return(result);
}

