#include "jamie.h"
#include "util.h"

/**********************************************************************
 * 
 * addlog
 *
 * Calculate addlog(a,b) = log[exp(a) + exp(b)]
 *
 * This makes use of the function log1p(x) = log(1+x) provided
 * in R's math library.   
 *
 **********************************************************************/
double addlog(double a, double b)
{
  if(b > a + THRESH) return(b);
  else if(a > b + THRESH) return(a);
  else return(a + log1p(exp(b-a)));
}

double sumlog(double *x, int n) {
  int i;
  double s=x[0];
  
  for(i=1; i<n; i++) 
    s = s + x[i];
  return(s);
}

/**********************************************************************
 * generate bernoulli RV with prob p
 **********************************************************************/
int rbern(double p) {
  double r = unif_rand();
  if(r>p)
    return(0);
  else 
    return(1);
}


/**********************************************************************  
 * sample_int 
 * Make a single draw from (1, ..., n) with probs (p_0, ..., p_(n-1))
 *
 **********************************************************************/
int sample_int(int n, double *p)
{
  int i;
  double r;

  /* R's random number generator */
  r = unif_rand();

  for(i=0; i<n; i++) {
    if(r < p[i]) return(i+1);
    else r -= p[i];
  }
  return(n); /* this shouldn't happen */
}

/* cumsum function, works for 2D array */
void cumsum2(double *lr, int nobs, int ncond, double *res) {
  int i,j;
  for(i=0; i<ncond; i++) {
    res[i*nobs] = lr[i*nobs];
    for(j=1; j<nobs; j++) {
      res[i*nobs+j] = res[i*nobs+j-1] + lr[i*nobs+j];
    }
  }
}

/* cumsum function, works for a vector */
void cumsum1(double *lr, int nobs, double *res) {
  int j;
  res[0] = lr[0];
  for(j=1; j<nobs; j++) {
    res[j] = res[j-1] + lr[j];
  }
}

/* compute sum of an array */
double dsum(double *y, int n) {
  double s=0;
  int i;
  for(i=0; i<n; i++)
    s=s+y[i];
  return(s);
}

